%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) 2015 Elisabeth Larsson
%
% DEMO1: Interpolate a function in 1-D using equispaced nodes,
% Gaussian RBFs and direct solution of the linear system arising from
% collocation at the node points. 
%
% Solve for different N and epsilon and plot the error.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear
close all
%
% The 1-D function to interpolate
%
f = @(x) 25./(25+(x-0.1).^2);

%
% Interval [a,b] over which to interpolate
%
a=-1; b=1;
%
% First solve for a fixed epsilon and different N.
% Spectral convergence is expected.
%
ep = 1;
N_vec = [5:35];
for k=1:length(N_vec)
  %
  % Uniform nodes
  %
  x = linspace(a,b,N_vec(k))';
  %
  % Compute distance matrix with r_ij = \|x_i-x_j\|
  %
  [cc,xx]=meshgrid(x);
  r = abs(xx-cc);
  %
  % Linear system  A*lambda=f, A(i,j)=phi(||x_i-x_j||)
  %
  A = exp(-ep^2*r.^2);
  lambda = A\f(x);
  %
  % Evaluate the interpolation error at 3*N points
  %
  xe = linspace(a,b,3*N_vec(k))';
  %
  % Compute distance matrix for evaluation points r_ij = \|xe_i-x_j\|
  %
  [cc,xx]=meshgrid(x,xe);
  re = abs(xx-cc);
  %
  % Evaluation matrix Ae(i,j) = phi(||xe_i-x_j||) 
  %
  Ae = exp(-ep^2*re.^2);
  fe = Ae*lambda;
  err = fe - f(xe);
  maxerr(k) = max(abs(err));
  maxlam(k) = max(lambda);
  minlam(k) = min(lambda);
end
%
% Plot the maximum norm error and the size of the coefficients lambda
%
figure(1),clf
subplot(2,2,1)
H=semilogy(N_vec,maxerr,'b-o');
fixplot('N','Max error',['\epsilon = ' num2str(ep)],H)

subplot(2,2,2)
H=plot(N_vec,sgnlog(maxlam),N_vec,sgnlog(minlam))
fixplot('N','\pm log_{10}(max/min(\lambda))',['\epsilon = ' num2str(ep)],H)

clear maxerr maxlam minlam

%---------------------------------------------------------------
%
% Now do the same thing for a fixed N and a range of epsilon-values
%
N=30;
ep_vec = logspace(-3,2,20);
%
% Uniform nodes
%
x = linspace(a,b,N)';
%
% Compute distance matrix with r_ij = \|x_i-x_j\|
%
[cc,xx]=meshgrid(x);
r = abs(xx-cc);
%
% We will evaluate the interpolation error at 3*N points
%
xe = linspace(a,b,3*N)';
%
% Compute distance matrix for evaluation points r_ij = \|xe_i-x_j\|
%
[cc,xx]=meshgrid(x,xe);
re = abs(xx-cc);

for k=1:length(ep_vec)  
  %
  % Linear system A*lambda=f, A(i,j)=phi(||x_i-x_j||)
  %
  A = exp(-ep_vec(k)^2*r.^2);
  lambda = A\f(x);
  %
  % Evaluation matrix Ae(i,j) = phi(||xe_i-x_j||) 
  %
  Ae = exp(-ep_vec(k)^2*re.^2);
  fe = Ae*lambda;
  err = fe - f(xe);
  maxerr(k) = max(abs(err));
  maxlam(k) = max(lambda);
  minlam(k) = min(lambda);
end
%
% Plot the maximum norm error and the size of the coefficients lambda
%
subplot(2,2,3)
H=loglog(ep_vec,maxerr,'b-o');
fixplot('\epsilon','Max error',['N = ' num2str(N)],H)

subplot(2,2,4)
H=semilogx(ep_vec,sign(maxlam).*log10(abs(maxlam)),ep_vec,sign(minlam).* ...
     log10(abs(minlam)))
fixplot('\epsilon','\pm log_{10}(max/min(\lambda))',['N = ' num2str(N)],H)
  
  


