%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) 2015 Elisabeth Larsson
%
% DEMO2: Numerically investigate the condition number of the
% interpolation matrix A, and the interpolation error for Gaussian RBFs,
% using scattered nodes in the unit disc, for the direct method as well
% as the RBF-QR method.  
%
% Compare the locations of blue and red regions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear
close all
%
% Function to interpolate
%
fun = @(x) 25./(25+(x(:,1)-0.2).^2+2*x(:,2).^2);
%
% Node sets for the unit disc will be loaded from file (the name of the
% node set variable is in all cases x)
%
addpath('./DataFiles/')
%
% Select a range of existing N-values 
%
%N=[3 4 5 6 9 10 11 12 13 14 15 16 20 21 23 24 25 28 29 30 31 32 33 34 35 ...
%   36 37 45 48 49 50 57 60 64 66 75];
N = [10 15 21 28 36 45 50 60 75];
%
% Variation with epsilon normally is best illustrated in log-scale
%
ep = logspace(-2.5,1,21);
%
% Choice of RBF, 'gs' means Gaussian, also see 'mq','iq',...
%
phi = 'gs';
%
% Evaluate all errors at 100 nodes in the unit disc
%
Ne = 100;
fname=['unitdisc' num2str(Ne)]; 
load(fname); xe = x;
fe = fun(xe);

warning off
for k=1:length(N)
  Current_N = N(k)
  fname=['unitdisc' num2str(N(k))]; 
  load(fname);
  %
  % The function xcdist computes a distance matrix. First argument
  % evaluation points and second argument nodes (centers)
  %
  r = xcdist(x,x);
  re = xcdist(xe,x);
  f = fun(x);
  
  for j=1:length(ep)
    %
    % RBFmat computes RBF-matrices. The '0' indictes derivative of order
    % zero. That is, the interpolation matrix.
    %
    A = RBFmat(phi,ep(j),r,'0');
    Ae = RBFmat(phi,ep(j),re,'0');
    %
    % NOTE that here we skip computing lambda. Ae/A is the 0th order
    % differentiation matrix or an evaluation matrix.
    %
    se = (Ae/A)*f;
    %
    condA(k,j) = cond(A);
    maxerr(k,j) = max(abs(se-fe));
    %
    % We will perform the same computation using RBF-QR. The method only
    % works for small epsilon, so above in this case ep=2, we replicate
    % the previous result.
    %
    if (ep(j)<2)
      %
      % RBF_QR_diffmat directly produces the differentiation
      % matrix. Unfortunatley, the interpolation operator is here called
      % '1' for unity (not '0'). The other arguments are the evaluation
      % points and the nodes (centers), and epsilon.
      %
      [Aqr,Psi] = RBF_QR_diffmat_2D('1',xe,x,ep(j));
      sqr = Aqr*f;
      errqr(k,j) = max(abs(sqr-fe));
      condqr(k,j) = cond(Psi.A0); % The interpolation matrix is a part of Psi
    else
      errqr(k,j)=maxerr(k,j);
      condqr(k,j) = condA(k,j);
    end
  end
end
warning on

log_ep=log10(ep);
log_cA=log10(condA);
%
% All of this is to make the plot look nice as the condition estimator
% cannot handle more than 10^20.
%
mask1 = log_cA > 17;
mask2 = log_cA <= 17;
log_cA=mask2.*log_cA + 18*mask1;
log_cA(1,end)=21;
v_c=[0:21];

figure(1),clf
[C,H1]=contourf(log_ep,N,log_cA,v_c);
H2=colorbar;
fixcontourplot('log_{10} (\epsilon)','N', ...
               'RBF-Direct: log_{10}( cond_2(A) )',H1,C,H2)
set(H2,'YTick',[0 5 10 15 18])
ax=caxis;

log_err=log10(maxerr);
v_e=-12:0.5:0;

figure(2),clf
[C,H1]=contourf(log_ep,N,log_err,v_e);
H2=colorbar;
fixcontourplot('log_{10} (\epsilon)','N', ...
               'RBF-Direct: log_{10} |s(x)-f(x)| ',H1,[],H2)



log_cqr=log10(condqr);

figure(3),clf
[C,H1]=contourf(log_ep,N,log_cqr,v_c);
H2=colorbar;
fixcontourplot('log_{10} (\epsilon)','N', ...
               'RBF-QR: log_{10}( cond_2(A) )',H1,C,H2)
caxis(ax)

log_eqr=log10(errqr);

figure(4),clf
[C,H1]=contourf(log_ep,N,log_eqr,v_e)
H2=colorbar;
fixcontourplot('log_{10} (\epsilon)','N', ...
               'RBF-QR: log_{10} |s(x)-f(x)| ',H1,[],H2)

