%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) 2015 Elisabeth Larsson
%
% DEMO5: Solve the Poisson equation in 2-D on a star shaped domain using
% Gaussian RBFs and both the direct approach and RBF-QR. You can try both
% quasi uniform and clustered nodes.
%
% Compare the results with and without RBF-QR and with different node
% sets. Also try the two different domains.   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear
close all
%
% The function to interpolate and its first derivatives
%
fun = @(x1,x2) sin(x1.^2+2*x2.^2) - sin(2*x1.^2+(x2-0.5).^2);
dfun1 = @(x1,x2) 2*x1.*cos(x1.^2 + 2*x2.^2) ...
        - 4*x1.*cos((x2 - 1/2).^2 + 2*x1.^2);
dfun2 = @(x1,x2) 4*x2.*cos(x1.^2 + 2*x2.^2) ...
        - cos((x2 - 1/2).^2 + 2*x1.^2).*(2*x2 - 1);
Lfun = @(x1,x2) 6*cos(x1.^2 + 2*x2.^2) - 6*cos((x2 - 1/2).^2 + 2*x1.^2) ...
       - 4*x1.^2.*sin(x1.^2 + 2*x2.^2) - 16*x2.^2.*sin(x1.^2 + 2*x2.^2) ...
       + sin((x2 - 1/2).^2 + 2*x1.^2).*(2*x2 - 1).^2 ...
       + 16*x1.^2.*sin((x2 - 1/2).^2 + 2*x1.^2);
%
% A star shaped domain has a boundary that can be expressed as r(theta)
%
% This is a flower domain
% rb = @(theta) 1+0.8*(cos(4*theta));
% R = 1.8; % The maximum radius
%
% This one we call a starfish
rb = @(theta) 1+0.1*(sin(6*theta)+sin(3*theta));
R = 1.2; %Actually less, this is an upper bound

%
% Use Halton nodes in the interior. Start from nodes in a square.
%
ndims=2;
N0 = 100;
x0 = R*(-1 + 2*halton(N0,ndims));
%
% Estimate the fraction of the square covered by the domain.
%
[th,r]=cart2pol(x0(:,1),x0(:,2));
in = find(r<rb(th));
Aomega = length(in)/N0;
r0 = xcdist(x0,x0);
%
% Average distance between nearest nodes
%
h0 = sum(min(r0+eye(N0)))/N0
%
% Use estimated arclength to distribute boundary nodes
%
theta0=0:2*pi/N0:2*pi;
r = rb(theta0);

x0 = r.*cos(theta0);
y0 = r.*sin(theta0);
%
% Measure the Euclidian length of each theta interval
%
l = sqrt(diff(x0).^2 + diff(y0).^2);
L = [0 cumsum(l)];
%
% L(theta) is a monotone function that can be inverted to give theta(L).
% That is, we can find theta values that give a uniform distribution of
% arclength between boundary nodes through
%

%
% Errors will be measured over u uniform polar grid.
%
Ne_t = 60;
Ne_r = 20;

th = 0:2*pi/Ne_t:2*pi;
r = 0:1/(Ne_r-1):1;
[rr,tt]=meshgrid(r,th);
rr = diag(rb(th))*rr;
[xx,yy] = pol2cart(tt,rr);
xe = [xx(:) yy(:)];
%[xe(:,1),xe(:,2)]=pol2cart(tt(:),rr(:));

mesh(rr,tt,0*rr)

figure(1),clf
mesh(rr.*cos(tt),rr.*sin(tt),0*rr)
view(2)
hold on
plot(x0,y0,'k-')

%
% Select approximate node numbers for the interpolation
% 
Nvec = [25 50 100 200 400 800];
ep = 0.5;
phi = 'gs';
for k=1:length(Nvec)
  N = Nvec(k)
  %
  % Generate interior nodes
  %
  x = R*(-1 + 2*halton(ceil(N/Aomega),ndims));
  [th,r]=cart2pol(x(:,1),x(:,2));
  in = find(r<rb(th));
  x = x(in,:);
  Nreal(k) = size(x,1)
  %
  % Generate nodes on the boundary. Estimate the distance between nodes
  % through scaling of h0.
  %
  h=h0*sqrt(N0/N)
  %
  % Divide the full arclength into pieces of size h
  %
  Nb = ceil(L(end)/h) + 1;
  lg = linspace(0,L(end),Nb);
  lg = lg(1:end-1)'; % Last point same as first
  Nb = Nb-1;
  % 
  % Interpolate the arclength function L(theta) to get theta for the
  % boundary points
  %
  thb = interp1(L,theta0,lg);
  xb = rb(thb).*cos(thb);
  yb = rb(thb).*sin(thb);
  %
  % Locate interior points that are too close to the boundary points.
  % How close is too close?
  % 
  rbc = xcdist([xb,yb],x) + 2*R*eye(length(xb),Nreal(k));
  xb = [xb,yb];
  bdist = min(rbc);
  pos = find(bdist>0.5*h);
  x = x(pos,:);
  x = [xb;x];
  Nreal(k) = size(x,1)
  %
  % Now, we are going to cluster nodes towards the boundary. We see each
  % point as lying on a ray of angle theta from the origin in the
  % interval [0,rb(theta)]. The nodes will be moved through a
  % Chebyshev-like transformation along each ray.
  %
  [th,r]=cart2pol(x(:,1),x(:,2));
  rbx = rb(th);
  r = rbx.*sin(0.5*pi*r./rbx);
  [x(:,1),x(:,2)]=pol2cart(th,r);
  %
  % Solve using RBF-Direct
  % 
  nd = 2;
  r = xcdist(x,x);
  re = xcdist(xe,x);
  A = RBFmat(phi,ep,r,'0');
  A_L = RBFmat(phi,ep,r(Nb+1:Nreal(k),:),'L',nd);
  I = eye(Nb,Nreal(k));
  D_L=[I;A_L/A];
  f = fun(xb(:,1),xb(:,2));
  g = Lfun(x(Nb+1:end,1),x(Nb+1:end,2));
  rhs = [f;g];
  %
  % u=D_L*f
  %
  u = D_L\rhs;
  Ae = RBFmat(phi,ep,re,'0');
  ue = (Ae/A)*u;
  %
  % Solve using RBF-QR
  %
  [A_L,Psi]=RBF_QR_diffmat_2D('L',x(Nb+1:end,:),x,ep);
  [Ae,Psi]=RBF_QR_diffmat_2D('1',xe,x,ep);
  D_L = [I;A_L];
  
  uqr = Ae*(D_L\rhs);
  %
  % Errors
  %
  fe = fun(xe(:,1),xe(:,2));
  err(:,k) = abs(ue-fe);
  eqr(:,k) = abs(uqr-fe);
  maxerr(k) = max(err(:,k));
  maxeqr(k) = max(eqr(:,k));
  
  
  %
  % Plot the RBF-QR errors to see if they look right.
  %
  figure(2),clf
  [C,H1]=contourf(xx,yy,reshape(log10(err(:,k)),size(xx)));
  H2 = colorbar;
  hold on
  H=plot(x0,y0,'k-'); set(H,'LineWidth',2)
  set(H1,'EdgeColor','none')
  fixcontourplot('x_1','x_2','|Error|',H1,[],H2)
  figure(3),clf
  [C,H1]=contourf(xx,yy,reshape(log10(eqr(:,k)),size(xx)));
  H2=colorbar;
  hold on
  H=plot(x0,y0,'k-'); set(H,'LineWidth',2)
  set(H1,'EdgeColor','none')
  fixcontourplot('x_1','x_2','|Error|',H1,[],H2)
  figure(4),clf
  H=plot(xb(:,1),xb(:,2),'r*',x(Nb:end,1),x(Nb:end,2),'bo');
  fixplot('x_1','x_2','Nodes',H)
  pause
end
figure(5)
H=semilogy(Nvec,maxerr,'r--o',Nvec,maxeqr,'k-s');
fixplot('N','Max error','RBF-Direct (dashed) and RBF-QR (solid)',H)

figure(6),clf
H1=mesh(xx,yy,reshape(uqr,size(xx)));
hold on
H=plot(x0,y0,'k-'); set(H,'LineWidth',2)
set(H1,'FaceColor','interp')
fixplot('x_1','x_2','Solution with RBF-QR',H)

