%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) 2015 Elisabeth Larsson
% Based on a code initially written by Alfa Herydono, later modified by 
% Victor Scherbakov and Elisabeth Larsson
%
% DEMO6: Solve the Poisson equation in 2-D on a star shaped domain using
% Gaussian RBFs and both the direct approach and RBF-QR. You can try both
% quasi uniform and clustered nodes.
%
% Compare the results with and without RBF-QR and with different node
% sets. Also try the two different domains.   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
close all
clear
%
% This is a 2-dimensional problem
%
nd = 2;
%
% The function to interpolate and its derivatives
%
fun = @(x1,x2) sin(x1.^2+2*x2.^2) - sin(2*x1.^2+(x2-0.5).^2);
dfun1 = @(x1,x2) 2*x1.*cos(x1.^2 + 2*x2.^2) ...
        - 4*x1.*cos((x2 - 1/2).^2 + 2*x1.^2);
dfun2 = @(x1,x2) 4*x2.*cos(x1.^2 + 2*x2.^2) ...
        - cos((x2 - 1/2).^2 + 2*x1.^2).*(2*x2 - 1);
Lfun = @(x1,x2) 6*cos(x1.^2 + 2*x2.^2) - 6*cos((x2 - 1/2).^2 + 2*x1.^2) ...
       - 4*x1.^2.*sin(x1.^2 + 2*x2.^2) - 16*x2.^2.*sin(x1.^2 + 2*x2.^2) ...
       + sin((x2 - 1/2).^2 + 2*x1.^2).*(2*x2 - 1).^2 ...
       + 16*x1.^2.*sin((x2 - 1/2).^2 + 2*x1.^2);
%
% A star shaped domain has a boundary that can be expressed as r(theta)
%
% This is a flower domain
% rb = @(theta) 1+0.8*(cos(4*theta));
% R = 1.8; % The maximum radius
%
% This one we call a starfish
rb = @(theta) 1+0.1*(sin(6*theta)+sin(3*theta));
R = 1.2; %Actually less, this is an upper bound
%
% Problem and method parameters to choose
% 
phi = 'gs'; % The RBF (the only choice for RBF-QR)
ep = 0.5; % The shape parameter of the RBFs
Nvec = [100 120 160 200 240 300 400 800 1600]; % The approximate number of nodes
Np1 = 5; % The number of patches in one dimension
gamma = 0.15; % The relative overlap
%
%---------SET UP THE INFORMATION NEEDED FOR NODE GENERATION-------------
%
% We use Halton nodes and we need to know what number to choose
%
N0 = 200; % Used just for testing the geometry
x0 = R*(-1 + 2*halton(N0,nd));
%
% Estimate the fraction of the square covered by the domain.
%
[th,r]=cart2pol(x0(:,1),x0(:,2));
in = find(r<rb(th));
Aomega = length(in)/N0;
r0 = xcdist(x0,x0);
%
% Average distance between nearest nodes
%
h0 = sum(min(r0+eye(N0)))/N0;
%
% We will use estimated arclength to distribute boundary nodes
%
theta0=0:2*pi/N0:2*pi;
r = rb(theta0);

x0 = r.*cos(theta0);
y0 = r.*sin(theta0);
%
% Measure the Euclidian length of each theta interval
%
l = sqrt(diff(x0).^2 + diff(y0).^2);
L = [0 cumsum(l)];
%
% L(theta) is a monotone function that can be inverted to give theta(L).
% That is, we can find theta values that give a uniform distribution of
% arclength between boundary nodes through
%
uniarcl = @(lb) interp1(L,theta0,lb);
%
%------------ CREATE AND PLOT THE EVAL GRID -----------------------------
%
% Errors will be measured over a uniform polar grid.
%
Ne_t = 60;
Ne_r = 20;

th = 0:2*pi/Ne_t:2*pi;
r = 0:1/(Ne_r-1):1;
[rr,tt]=meshgrid(r,th);
rr = diag(rb(th))*rr;
[xx,yy] = pol2cart(tt,rr);
xe = [xx(:) yy(:)];
Ne = size(xe,1);

figure(1),clf
mesh(rr.*cos(tt),rr.*sin(tt),0*rr)
view(2)
hold on
plot(x0,y0,'k-')
% 
%------------ CREATE PATCHES THAT COVER THE DOMAIN------------------
%
% Start with non-overlapping boxes.
% Note that for the flower domain, Np must be odd for success or large
% enough so that the narrow middle part is resolved.
%
R0 = 2*R/(2*Np1); % Half size of box side
Rp = sqrt(2)*R0*(1+gamma); % Radius of circle with overlap
%
% The centers of the preliminary layout are on a grid of size 2*R0
%
xp = (-R+R0):2*R0:(R-R0);
[xp,yp]=meshgrid(xp);
Cp = [xp(:) yp(:)];
%
% If the center of a patch is outside the domain, then discard it.
% Note: This is something that may actually go wrong if the patches are
% too large.
%
[thp,rp]=cart2pol(Cp(:,1),Cp(:,2));
pos = find(rb(thp) > rp);
Cp = Cp(pos,:);
Rpi = Rp*ones(size(Cp(:,1)));
%
% Boundary points that are not covered will be included in the closest
% circle. The circle then needs to be enlarged.
%
r0x = rb(theta0).*cos(theta0);
r0y = rb(theta0).*sin(theta0);
hb = max(sqrt(diff(r0x).^2+diff(r0y).^2));
for k=1:size(Cp,1)
  dist2(:,k) = sqrt((r0x(:)-Cp(k,1)).^2 + (r0y(:)-Cp(k,2)).^2);
end
[rbc,circind] = min(dist2,[],2);
for k=1:size(Cp,1)
  pos = find(circind==k)
  if (length(pos)>0)
    rk = max(rbc(pos));
    Rpi(k) = max(Rp,rk+hb/2); % Extra margin to cover pts in between
  end
end
%
% Create a patch struct to store the patch info
%
Np = size(Cp,1);
midpt = num2cell(Cp,2);
[ptch(1:Np).midpt] = deal(midpt{:}); 
Rpi = num2cell(Rpi,Np);
[ptch(:).R] = deal(Rpi{:});
%
% Draw the circles
%
rx = cos(theta0);
ry = sin(theta0);
for k= 1:Np
  plot(ptch(k).midpt(1) + ptch(k).R*rx,ptch(k).midpt(2)+ptch(k).R*ry,'r-')
  text(ptch(k).midpt(1),ptch(k).midpt(2),num2str(k))
end
%
%--------------FOR EACH NUMBER OF NODES SOLVE THE PDE-----------------
%
for k=1:length(Nvec)
  N = Nvec(k)
  %
  % Generate interior nodes
  %
  x = R*(-1 + 2*halton(ceil(N/Aomega),nd));
  [th,r]=cart2pol(x(:,1),x(:,2));
  in = find(r<rb(th));
  x = x(in,:);
  Nreal(k) = size(x,1);
  N = Nreal(k);
  %
  % Generate nodes on the boundary. Estimate the distance between nodes
  % through scaling of h0.
  %
  h=h0*sqrt(N0/N)
  %
  % Divide the full arclength into pieces of size h
  %
  Nb = ceil(L(end)/h) + 1;
  lg = linspace(0,L(end),Nb);
  lg = lg(1:end-1)'; % Last point same as first
  Nb = Nb-1;
  % 
  % Interpolate the arclength function L(theta) to get theta for the
  % boundary points
  %
  thb = uniarcl(lg);
  xb = rb(thb).*cos(thb);
  yb = rb(thb).*sin(thb);
  %
  % Locate interior points that are too close to the boundary points.
  % How close is too close?
  % 
  xb = [xb,yb];
  rbc = xcdist(xb,x) + 2*R*eye(size(xb,1),N);
  bdist = min(rbc);
  pos = find(bdist>0.5*h);
  x = x(pos,:);
  xc = [xb;x];
  Nreal(k) = size(xc,1);
  N = Nreal(k)
  %
  % Now, we are going to cluster nodes towards the boundary. We see each
  % point as lying on a ray of angle theta from the origin in the
  % interval [0,rb(theta)]. The nodes will be moved through a
  % Chebyshev-like transformation along each ray.
  %
  [th,r]=cart2pol(xc(:,1),xc(:,2));
  rbx = rb(th);
  r = rbx.*sin(0.5*pi*r./rbx);
  [xc(:,1),xc(:,2)]=pol2cart(th,r);
  
  % Organize RBF nodes using structure 
  % Basically find which nodes fall in which partitions and store this 
  % information in the structure
  %xc = num2cell(xc,2);
  %Nc = length(xc(:,1));
  %[data(1:Nc).ctr] = deal(xc{:}); 
  %[data.circind] = deal([]);
  %xc = cat(1,xc{:});
%
% Not sure if we need circind, wait and see
%


for i=1:Np
    flagin = hypot(ptch(i).midpt(1)-xc(:,1),ptch(i).midpt(2)-xc(:,2)) < ptch(i).R;
    locpts(i).ind = find(flagin);   
    Ni(i)=length(locpts(i).ind);
    flagin = hypot(ptch(i).midpt(1)-xe(:,1),ptch(i).midpt(2)-xe(:,2)) < ptch(i).R;
    elocpts(i).ind = find(flagin); 
    Nei(i)=length(elocpts(i).ind);
    %    for j=1:length(ptch(i).indlocpts)
    %    data(ptch(i).indlocpts(j)).circind = [data(ptch(i).indlocpts(j)).circind i];
    %end
end

% Define the partition of unity weight using Shepard's method
% We do not want to recompute s everytime. Do all at once.
  %
  % Update the call if possible. Strange sequence.
  % 
    pu = puweightrbf(xc,locpts,ptch);
    epu = puweightrbf(xe,elocpts,ptch);
    
    wplot = zeros(Ne,1);
    wplot(elocpts(1).ind) = epu(1).w;

    %% RBF matrices

    B = spalloc(N,N,sum(Ni.^2));
    Bqr = spalloc(N,N,sum(Ni.^2));
    for i=1:Np
      %   
      % Build local RBF matrices using RBF-Direct 
      %
      ind = locpts(i).ind;
      all = 1; % To compute directional derivatives from xcdist 
      rc = xcdist(xc(ind,:),xc(ind,:),all);
      A    = RBFmat(phi,ep,rc,'0');
      A_L  = RBFmat(phi,ep,rc,'L',nd); % Laplacian
      Ax  = RBFmat(phi,ep,rc,'1',1); % First derivative in first dim
      Ay  = RBFmat(phi,ep,rc,'1',2); % First derivative in secnd dim

      % Construct derivative operator matrices
      D0 = diag(pu(i).w)*A; 
      Dx = diag(pu(i).w)*Ax + diag(pu(i).wx)*A;
      Dy = diag(pu(i).w)*Ay + diag(pu(i).wy)*A;
      
      D_L = diag(pu(i).w)*A_L + ...
            2*diag(pu(i).wx)*Ax + 2*diag(pu(i).wy)*Ay + ...
            diag(pu(i).wxx + pu(i).wyy)*A;
      %
      % Add the local differentiation matrix into the global matrix
      %
      B(ind,ind) = B(ind,ind) + D_L/A;
    
      % Build the evaluation matrix
      ind_e = elocpts(i).ind;
      if (length(ind_e)>0)
      re = xcdist(xe(ind_e,:),xc(ind,:));
      E(i).Ae  = RBFmat(phi,ep,re,'0');
      E(i).Ae = diag(epu(i).w)*E(i).Ae/A;
      else
        E(i).Ae=zeros(0,length(ind));
      end
      %   
      % Build local RBF matrices using RBF-QR
      %
      [A_Lqr,Psi] = RBF_QR_diffmat_2D('L',xc(ind,:),xc(ind,:),ep);
      [Axqr,Psi]  = RBF_QR_diffmat_2D('x',xc(ind,:),Psi);
      [Ayqr,Psi]  = RBF_QR_diffmat_2D('y',xc(ind,:),Psi); 
      I = eye(length(ind));
      % Construct derivative operator matrices
      size(pu(i).w),size(I)
      D0 = diag(pu(i).w)*I; 
      Dx = diag(pu(i).w)*Axqr + diag(pu(i).wx)*I;
      Dy = diag(pu(i).w)*Ayqr + diag(pu(i).wy)*I;
      
      D_Lqr = diag(pu(i).w)*A_Lqr + ...
            2*diag(pu(i).wx)*Axqr + 2*diag(pu(i).wy)*Ayqr + ...
            diag(pu(i).wxx + pu(i).wyy)*I;
      %
      % Add the local differentiation matrix into the global matrix
      %
      Bqr(ind,ind) = Bqr(ind,ind) + D_Lqr;
    
      % Build the evaluation matrix
      ind_e = elocpts(i).ind;
      if (length(ind_e)>0)
        re = xcdist(xe(ind_e,:),xc(ind,:));
        [Eqr(i).Ae,Psi]  = RBF_QR_diffmat_2D('1',xe(ind_e,:),Psi);
        Eqr(i).Ae = diag(epu(i).w)*Eqr(i).Ae;
      else
        Eqr(i).Ae=zeros(0,length(ind));
      end  
    end
    %
    % Modify the equations corresponding to the boundary conditions
    %
    B(1:Nb,1:N) = eye(Nb,N);
    Bqr(1:Nb,1:N) = eye(Nb,N);
    %
    % Right hand side
    %
    f = fun(xb(:,1),xb(:,2));
    g = Lfun(xc(Nb+1:end,1),xc(Nb+1:end,2));
    rhs = [f;g];
    %
    % Solve the system
    %
    u = B\rhs;
    uqr = Bqr\rhs;
    %
    % Evaluate and plot
    %
    ue  = zeros(Ne,1);
    uqre  = zeros(Ne,1);
    for i=1:Np
      ind=locpts(i).ind;
      ind_e = elocpts(i).ind;
      ue(ind_e) = ue(ind_e) + E(i).Ae*u(ind);
      uqre(ind_e) = uqre(ind_e) + Eqr(i).Ae*uqr(ind);
    end  
    fe = fun(xe(:,1),xe(:,2));
    err(:,k) = abs(ue-fe);
    eqr(:,k) = abs(uqre-fe);
    maxerr(k) = max(err(:,k));
    maxeqr(k) = max(eqr(:,k));
    figure(2),clf
    [C,H1]=contourf(xx,yy,reshape(log10(err(:,k)),size(xx)));
    H2 = colorbar;
    hold on
    H=plot(x0,y0,'k-'); set(H,'LineWidth',2)
    set(H1,'EdgeColor','none')
    fixcontourplot('x_1','x_2','|Error|',H1,[],H2)
    figure(3),clf
    [C,H1]=contourf(xx,yy,reshape(log10(eqr(:,k)),size(xx)));
    H2=colorbar;
    hold on
    H=plot(x0,y0,'k-'); set(H,'LineWidth',2)
    set(H1,'EdgeColor','none')
    fixcontourplot('x_1','x_2','|Error|',H1,[],H2)
    pause
end
figure(5)
H=semilogy(Nvec,maxerr,'r--o',Nvec,maxeqr,'k-s');
fixplot('N','Max error','RBF-Direct (dashed) and RBF-QR (solid)',H)

figure(6),clf
H1=mesh(xx,yy,reshape(uqre,size(xx)));
hold on
H=plot(x0,y0,'k-'); set(H,'LineWidth',2)
set(H1,'FaceColor','interp')
fixplot('x_1','x_2','Solution with RBF-QR',H)


  

