%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) 2015 Elisabeth Larsson
% Based on codes initially written by Alfa Herydono and Elisabeth
% Larsson, later modified by 
% Victor Scherbakov and Elisabeth Larsson
%
% DEMO7: Solve the Heat equation in 2-D on a disc using
% Gaussian RBFs, RBF-PUM and RBF-QR.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
close all
clear
%
% This is a 2-dimensional problem
%
nd = 2;
%
% The initial condition. The boundary will be kept constant at this level
%
fun = @(x1,x2) exp(-x1.^2-2*x2.^2);
%
% Problem and method parameters to choose
% 
phi = 'gs'; % The RBF (the only choice for RBF-QR)
ep = 0.5; % The shape parameter of the RBFs
N = 400; % The approximate number of nodes
Np1 = 2; % The number of patches in one dimension
gamma = 0.1; % The relative overlap
T = 1; % The final simulation time
M = 100; % The number of time steps
R = 1; % Because of the unit disc
%
% Load the nodes. Unstructured, but quite uniform
%
N=100;
fname=['unitdisc' num2str(N)]; 
load(fname);
xc = x;
%
% The nodes are organized such that the boundary nodes come first.
%
r=sqrt(sum(x.^2,2));
pos = find(abs(r-1)<1e-6);
Nb = length(pos);
xb = xc(1:Nb,:);
%
% The boundary curve
%
N0=200;
theta0=0:2*pi/N0:2*pi;
x0 = cos(theta0);
y0 = sin(theta0);
%
%------------ CREATE AND PLOT THE EVAL GRID -----------------------------
%
% Errors will be measured over a uniform polar grid.
%
Ne_t = 60;
Ne_r = 20;

th = 0:2*pi/Ne_t:2*pi;
r = 0:1/(Ne_r-1):1;
[rr,tt]=meshgrid(r,th);
[xx,yy] = pol2cart(tt,rr);
xe = [xx(:) yy(:)];
Ne = size(xe,1);

figure(1),clf
mesh(rr.*cos(tt),rr.*sin(tt),0*rr)
view(2)
hold on
plot(x0,y0,'k-')
% 
%------------ CREATE PATCHES THAT COVER THE DOMAIN------------------
%
% Start with non-overlapping boxes.
% Note that for the flower domain, Np must be odd for success or large
% enough so that the narrow middle part is resolved.
%
R0 = 2*R/(2*Np1); % Half size of box side
Rp = sqrt(2)*R0*(1+gamma); % Radius of circle with overlap
%
% The centers of the preliminary layout are on a grid of size 2*R0
%
xp = (-R+R0):2*R0:(R-R0);
[xp,yp]=meshgrid(xp);
Cp = [xp(:) yp(:)];
%
% If the center of a patch is outside the domain, then discard it.
% Note: This is something that may actually go wrong if the patches are
% too large.
%
[thp,rp]=cart2pol(Cp(:,1),Cp(:,2));
pos = find(R > rp);
Cp = Cp(pos,:);
Rpi = Rp*ones(size(Cp(:,1)));
%
% Boundary points that are not covered will be included in the closest
% circle. The circle then needs to be enlarged.
%
hb = max(sqrt(diff(x0).^2+diff(y0).^2));
for k=1:size(Cp,1)
  dist2(:,k) = sqrt((x0(:)-Cp(k,1)).^2 + (y0(:)-Cp(k,2)).^2);
end
[rbc,circind] = min(dist2,[],2);
for k=1:size(Cp,1)
  pos = find(circind==k);
  if (length(pos)>0)
    rk = max(rbc(pos));
    Rpi(k) = max(Rp,rk+hb/2); % Extra margin to cover pts in between
  end
end
%
% Create a patch struct to store the patch info
%
Np = size(Cp,1);
midpt = num2cell(Cp,2);
[ptch(1:Np).midpt] = deal(midpt{:}); 
Rpi = num2cell(Rpi,Np);
[ptch(:).R] = deal(Rpi{:});
%
% Draw the circles
%
rx = cos(theta0);
ry = sin(theta0);
for k= 1:Np
  plot(ptch(k).midpt(1) + ptch(k).R*rx,ptch(k).midpt(2)+ptch(k).R*ry,'r-')
  text(ptch(k).midpt(1),ptch(k).midpt(2),num2str(k))
end
%
% Find the nodes that are in each patch  
%
for i=1:Np
    flagin = hypot(ptch(i).midpt(1)-xc(:,1),ptch(i).midpt(2)-xc(:,2)) < ptch(i).R;
    locpts(i).ind = find(flagin);   
    Ni(i)=length(locpts(i).ind);
    flagin = hypot(ptch(i).midpt(1)-xe(:,1),ptch(i).midpt(2)-xe(:,2)) < ptch(i).R;
    elocpts(i).ind = find(flagin); 
    Nei(i)=length(elocpts(i).ind);
    %    for j=1:length(ptch(i).indlocpts)
    %    data(ptch(i).indlocpts(j)).circind = [data(ptch(i).indlocpts(j)).circind i];
    %end
end

% Define the partition of unity weight using Shepard's method
% We do not want to recompute s everytime. Do all at once.
  %
  % Update the call if possible. Strange sequence.
  % 
    pu = puweightrbf(xc,locpts,ptch);
    epu = puweightrbf(xe,elocpts,ptch);
    

    %% RBF matrices

    Bqr = spalloc(N,N,sum(Ni.^2));
    for i=1:Np
      %   
      % Build local RBF matrices using RBF-Direct 
      %
      ind = locpts(i).ind;
      %   
      % Build local RBF matrices using RBF-QR
      %
      [A_Lqr,Psi] = RBF_QR_diffmat_2D('L',xc(ind,:),xc(ind,:),ep);
      [Axqr,Psi]  = RBF_QR_diffmat_2D('x',xc(ind,:),Psi);
      [Ayqr,Psi]  = RBF_QR_diffmat_2D('y',xc(ind,:),Psi); 
      I = eye(length(ind));
      % Construct derivative operator matrices
      D0 = diag(pu(i).w)*I; 
      Dx = diag(pu(i).w)*Axqr + diag(pu(i).wx)*I;
      Dy = diag(pu(i).w)*Ayqr + diag(pu(i).wy)*I;
      
      D_Lqr = diag(pu(i).w)*A_Lqr + ...
            2*diag(pu(i).wx)*Axqr + 2*diag(pu(i).wy)*Ayqr + ...
            diag(pu(i).wxx + pu(i).wyy)*I;
      %
      % Add the local differentiation matrix into the global matrix
      %
      Bqr(ind,ind) = Bqr(ind,ind) + D_Lqr;
    
      % Build the evaluation matrix
      ind_e = elocpts(i).ind;
      if (length(ind_e)>0)
        [Eqr(i).Ae,Psi]  = RBF_QR_diffmat_2D('1',xe(ind_e,:),Psi);
        Eqr(i).Ae = diag(epu(i).w)*Eqr(i).Ae;
      else
        Eqr(i).Ae=zeros(0,length(ind));
      end  
    end
    %
    % Find the time steps and coefficients for the BDF-2 method
    %
    [k,beta0,beta1,beta2]=BDF2coeffs(T,M);
    %
    % The time-stepping matrix is C=I-beta0*B for the interior
    % and I for the boundary
    %
    Cqr = speye(N);
    Cqr(Nb+1:end,:) = Cqr(Nb+1:end,:)-beta0*Bqr(Nb+1:end,:);
    [Lc,Uc] = lu(Cqr);
    %
    % Initial condition
    %
    u0 = fun(xc(:,1),xc(:,2));
    ue  = zeros(Ne,1);
    for i=1:Np
      ind=locpts(i).ind;
      ind_e = elocpts(i).ind;
      ue(ind_e) = ue(ind_e) + Eqr(i).Ae*u0(ind);
    end  
    %
    % Right hand side with boundary conditions
    % 
    rhs(:,1) = u0;
    rhs(1:Nb) = fun(xb(:,1),xb(:,2)); % Stays the same
    tn = k(1);
    %
    % Here comes the time stepping loop
    %
    for n = 1:M
      %
      % Solve the system to get solution at new time level
      %
      u = Uc\(Lc\rhs);
      %
      % Prepare the right hand side for the next step
      %
      nextstep = min(M,n+1);
      tn = sum(k(1:nextstep)); % Should be one step ahead        
      rhs = beta1(nextstep)*u-beta2(nextstep)*u0;
      % boundary conditions
      rhs(1:Nb) = fun(xb(:,1),xb(:,2)); % Stays the same;
      u0 = u;
      %
      % Evaluate and plot
      %
      ue  = zeros(Ne,1);
      for i=1:Np
        ind=locpts(i).ind;
        ind_e = elocpts(i).ind;
        ue(ind_e) = ue(ind_e) + Eqr(i).Ae*u(ind);
      end  
      figure(2),clf
      H1=mesh(xx,yy,reshape(ue,size(xx)));
      hold on
      H=plot(x0,y0,'k-'); set(H,'LineWidth',2)
      set(H1,'FaceColor','interp')
      fixplot('x_1','x_2', ...
              ['Solution at time t=' num2str(sum(k(1:nextstep-1)))],H)
      pause(0.1)
    end
        

